/*
  0.17: cleared code from 30000bytes to 25320 bytes, 16 Tracks now
  0.18: changed internal clock to interrupt routine
  0.19: switched setup to getDataValue() and fixed edit of multiple tracks
  0.20: redid the write and play stuff
  0.21: redid EEPROM organization
  0.22: fixed transpose finished play mode; included transpose and velocity tracks
  0.23: switched Potentiometer handling to new concept
  0.25: global transpose can use transpose track
  0.26: Trackpreparation after offTrigger of former step to improve the timing.
  0.27: MIDI Clock-Send and DIN-Sync/VMC-Clock, Global DividerDivider,  MIDI ClockReceive
  0.28: Some minor bugfixing.
  ToDo
   Split Menus in Track-Settings and Global-Settings and Save/Load
   Add separate Save/Load all track configs
   Add separate Save/Load all tracks stepvalues.
   Add separate Save/Load single tracks.
*/


#include <TimerOne.h>
#include <EEPROM.h>
#include <MIDI.h>
#include <LiquidCrystal.h>
MIDI_CREATE_DEFAULT_INSTANCE();
LiquidCrystal lcd(2, 3, 4, 5, 6, 7);

// Define inputs and outputs
#define button1Pin 8 //PLAY
#define button2Pin 9 //STOP
#define button3Pin 12 //WRITE
#define button4Pin 11 //F1
#define button5Pin 10 //F2
#define clockInPin 13 //CLOCK-IN
#define TempoCVPin A0 // =Tempo potentiometer
#define CV1Pin A1 // (5V or CV1-In) via CV1 potentiometer
#define CV2Pin A2 // (5V or CV2-In) via CV2 potentiometer


#define dinClockPin  A3 //VMC-BUS
#define dinGatePin  A4 //VMC-BUS
#define clockOutPin A5 //->sent via the clock-socket-switch to clockInPin

#define input0 A0 // =Tempo potentiometer
#define input1 A1 // (5V or CV1-In) via CV1 potentiometer
#define input2 A2 // (5V or CV2-In) via CV2 potentiometer
uint16_t input[3];



// Variables to handle the potentiometers
uint16_t invalueRaw[3]; /* Used in function poti adjust */
uint16_t invalueOffset[3]; /* The difference between 1023 and the real AD-ed value */
uint16_t adoptedRead[3]; /* The calculated value that makes 1023 from the real AD-ed value */
uint8_t invalue[3]; /* The value 0-127 calculated from adoptedRead[8] using invalueOffset[8] */
uint8_t invaluePrev[3];
unsigned long lastRead[3]; /* used for correcting false retriggering*/
unsigned long thisRead[3]; /* used for correcting false retriggering*/
unsigned long CV[3];

/*
  //Offset variables for the CV ins, for adopting to potentiometer variations
  unsigned long CV1raw;
  unsigned long CV2raw;
  unsigned long TempoCVraw;

  //Offset variables for the CV ins, for adopting to potentiometer variations
  unsigned long maxOffTempoCV;
  unsigned long maxOffCV1;
  unsigned long maxOffCV2;

  // Variables to handle the potentiometers
  unsigned long CV1;
  unsigned long CV2;
  unsigned long TempoCV;
*/


// Variables to handle the buttons and digital inputs
byte button1State = 0;
byte button2State = 0;
byte button3State = 0;
byte button4State = 0;
byte button5State = 0;
byte clockInState = 0;
byte button1StatePrev = 0;
byte button2StatePrev = 0;
byte button3StatePrev = 0;
byte button4StatePrev = 0;
byte button5StatePrev = 0;
byte clockInStatePrev = 0;

// Variable required for the clock generator
uint8_t tempo;
long period;
uint16_t midiclockcounter = 0; // Counts incoming MIDI-Clock or DIN-Sync clock related to the divider settings
uint16_t intclockcounter = 0; // Counts the internal clock related to the divider settings
uint16_t dinclockcounter = 0; // Counts the internal clock for 24ppq for DIN-Sync out and MIDI-clock out

unsigned long clockfactor = 1;
byte startFlag = 0;
byte quarterCount = 0;
byte barCount = 0;


// Variables for systemwider setup parameters

byte inchannel = 1;

byte notelength = 1;
byte transposemode = 3;
byte transposecenter = 60;
byte noteoffmode = 0;
byte velocitymode = 1;
byte notethrumode = 3;
byte controlthrumode = 1;
byte programthrumode = 0;
byte bendthrumode = 1;
byte otherthrumode = 0;
byte chordtriggermode = 1;
byte clockmode = 0;
byte velocitymod = 0;

/* Variable used for velocity control */
byte center;
byte range;

// Variables for track data
const byte maxSteps = 16;
const byte maxTracks = 16;
//byte stepvalue[maxSteps][maxTracks];
byte stepvalue[maxSteps][maxTracks];
byte noteOff[maxTracks];
byte ccnumber[maxTracks];
byte ccnumberall = 7;
byte clockdivider[maxTracks];
byte inclockdivider = 0;
byte clockcounter[maxTracks];
byte laststep[maxTracks];
byte outchannel[maxTracks];
byte outchannelall = 10;
uint8_t stepnumber[maxTracks];
byte tracktype[maxTracks];
byte tracktypeall = 0;
int transpose[maxTracks];
int transposeall = 0;
byte transposesource[maxTracks];
byte velocity[maxTracks];
byte velocityall = 64;
byte velocitysource[maxTracks];
byte currentvalue[maxTracks];
byte osmoseenable = 0;
byte sendclock = 0;
byte startstopmode = 0;


// Variables for sequence handling
byte playedtrack; // used like a normal for i in ... counter, but the name helps understanding the code
byte selectedtrack;
byte pause[maxTracks]; // 0=normal play, 1=steps are counted, but no midi is sent, 2=will play again after next cycle
//byte waitEndOfPause[maxTracks]; // should be replaced by pause[track]==2
byte playonstart[maxTracks];
byte writeStepNumber = 0;
byte stepBack = 0;
byte stepLCDposition;
byte ontrigger[maxTracks];
byte offtrigger[maxTracks];
byte i = 0;
byte j = 0;

// Variables to handle the generic status of the software
byte setupmode = 0;
byte menu = 0;
byte menuPrev = 0;
byte edit = 0;
byte runmode = 0; // 0=Stop, 1=Play
byte writemode = 0; // 1=Record Trigger, 2=Record Velocity
byte writeselect = 0;
byte adjuststep = 0; //used for the poti adjustment procedure
byte normaltrack;
byte setuptrack;
byte myvalue;
byte maxValue;
byte minValue;
byte incten;
byte runover;
byte saveMemory;
byte saveValue;
byte saveValueMulti[maxTracks];
byte saveMulti; // when leavin the edit mode 1 = this menu requires saving the parameters of all tracks
byte displaymode = 0;




void setup() {

  invalueOffset[0] = EEPROM.read(1);
  invalueOffset[1] = EEPROM.read(2);
  invalueOffset[2] = EEPROM.read(3);


  // put your setup code here, to run once:
  lcd.begin(16, 2);
  MIDI.begin();
  MIDI.setInputChannel(inchannel);
  if (otherthrumode == 0) {
    MIDI.setThruFilterMode(0);
    MIDI.turnThruOff();
  }
  if (otherthrumode == 1) {
    MIDI.setThruFilterMode(3);
    MIDI.turnThruOn(3);
  }

  pinMode(button1Pin, INPUT);
  pinMode(button2Pin, INPUT);
  pinMode(button3Pin, INPUT);
  pinMode(button4Pin, INPUT);
  pinMode(button5Pin, INPUT);
  pinMode(clockInPin, INPUT);
  pinMode(input0, INPUT);
  pinMode(input1, INPUT);
  pinMode(input2, INPUT);
  pinMode(clockOutPin, OUTPUT);

  input[0] = input0;
  input[1] = input1;
  input[2] = input2;
  tempo = 120;
  /* TIMER FOR THE INTERNAL CLOCK */
  Timer1.initialize(100000);
  Timer1.attachInterrupt( timerIsr );
  period = (312500 / tempo);
  Timer1.setPeriod(period);

  transposecenter = EEPROM.read(7);

  /* FIRMWARE VERSION */
  lcd.setCursor(0, 0);
  lcd.print(F("SQ-4m V0.28     "));
  lcd.setCursor(0, 1);
  lcd.print(F("                "));
  triggerreset();
  loadsequence();
  clearscreen();
  loadsettings();
  clockmode = 0;
  setClockFactor();
  menu=16;
}

/* ------ FULL SYSTEM RESET ------- */
void(* resetFunc)  (void) = 0;



void loop() {
  /* DEBUGGING OUTPUTS - COMMENTED NORMALLY */
  /*
    lcd.setCursor(0, 1);
    lcd.print(button4State);
    lcd.print(button4StatePrev);
    lcd.print(button5State);
    lcd.print(button5StatePrev);
    //*/



  /* READ BUTTONS */
  button1State = !digitalRead(button1Pin);
  button2State = !digitalRead(button2Pin);
  button3State = !digitalRead(button3Pin);
  button4State = !digitalRead(button4Pin);
  button5State = !digitalRead(button5Pin);
  for (i = 0 ; i <= 2 ; i++) {
    lastRead[i] = thisRead[i];
    thisRead[i] = analogRead(input[i]);
    invaluePrev[i] = invalue[i];
    adoptedRead[i] = int(thisRead[i] + ((thisRead[i] + 1.0) / (1023.0 - (invalueOffset[i] + 1) )) * invalueOffset[i]);
    if (adoptedRead[i] > 1023) adoptedRead[i] = 1023;
    if (i == 0) {
      invalue[i] = int(adoptedRead[i] / 4);
      if (invalue[i] > 255) invalue[i] = 255;
      if (invalue[i] < 1) invalue[i] = 1;
    }
    if (i > 0) {
      invalue[i] = int(adoptedRead[i] / 8);
      if (invalue[i] > 127) invalue[i] = 127;
    }
    if (invalue[i] < 0) invalue[i] = 0;
  }

  center = (invalue[1]);
  range = (invalue[2]);

  /* ------------ FULL SYSTEM RESET WHEN ALL BUTTONS ARE PRESSED ------------ */
  if (button1State == 1 && button2State == 1 && button3State == 1 && button4State == 1 && button5State == 1) {
    //resetFunc();
  }

  if ( invalue[0] != invaluePrev[0]) { // A change at the Tempo potentiometer sets a new tempo
    //period = (312500 /invalue[0]);
    if (invalue[0] > 0 && invalue[0] < 256) tempo = invalue[0];
    Timer1.setPeriod(312500 / tempo);
    if (displaymode == 0 && transpose == 0) {
      lcd.setCursor(5, 0);
      lcd.print(tempo);
      lcd.print(F(" "));
    }
  }




  /* BUTTONS IN THE NORMAL MENUS */


  if (setupmode == 0) {
    /* HOLD  WRITE AND PRESS STOP DELETES THE PATTERN */
    if (button2State == 0 && button2StatePrev == 1 && button3State == 1) deletesequence();

    /* HOLD  STOP AND PRESS WRITE SENDS MIDI-RESET */
    if (button3State == 0 && button3StatePrev == 1 && button2State == 1) midireset();

    /* TIE plus REST will switch to setup in Playmode and Stopmode*/
    if (writemode == 0) {
      if (button4State == 1 && button4StatePrev == 0 && button5State == 1) {
        button4StatePrev = 1;
        setupmode = 1;
        clearscreen();
      }
      if (button4State == 0 && button4StatePrev == 1) {
        button4StatePrev = 0;
      }
    }




    /* PRINT PLAY OR STOP TO DISPLAY */
    lcd.setCursor(0, 0);
    if (runmode == 0 && writemode == 0) {
      lcd.setCursor(0, 0);
      lcd.print(F("STOP"));
    }
    if (runmode == 1  && pause[selectedtrack] == 0 ) {
      lcd.setCursor(0, 0);
      lcd.print(F("PLAY"));
    }
    if (runmode == 1  && pause[selectedtrack] == 1) {
      lcd.setCursor(0, 0);
      lcd.print(F("play"));
    }
    if (runmode == 1  && pause[selectedtrack] == 2) {
      lcd.setCursor(0, 0);
      lcd.print(F("plAY"));
    }



    if (writemode == 0) {
      /* TIE plus REST will switch to setup in Playmode and Stopmode*/
      if (button4State == 1 && button5State == 1) {
        //normaltrack = selectedtrack;
        //selectedtrack = setuptrack;
        setupmode = 1;
        clearscreen();
      }

      /* F2 PRESSED + BUTTON START/STOP SELECTS TRACK */
      if (button5State == 1 && button1State == 1 && button1StatePrev == 0 && writemode == 0) {
        if (selectedtrack == 0) selectedtrack = maxTracks;
        selectedtrack--;
        button1StatePrev = 1;
      }
      if (button1State == 0 && button1StatePrev == 1) button1StatePrev = 0;

      if (button5State == 1 && button2State == 1 && button2StatePrev == 0 && writemode == 0) {
        selectedtrack++;
        if (selectedtrack == maxTracks) selectedtrack = 0;
        button2StatePrev = 1;
      }
      if (button2State == 0 && button2StatePrev == 1) button2StatePrev = 0;

      lcd.setCursor(12, 0);

      lcd.print(F("Tr"));
      lcd.print(selectedtrack + 1);
      lcd.print(F(" "));
    }

    /* F1 PRESSED + STOP MUTES ALL TRACKS*/
    if (button4State == 1 && button2State == 1 && button2StatePrev == 0 && writemode == 0) {
      for (i = 0; i <= (maxTracks - 1); i++) {
        pause[i] = 1;
      }
      button2StatePrev = 1;
      lcd.setCursor(0, 1);
      lcd.print(F("all tracks muted"));
    }
    if (button2State == 0 && button2StatePrev == 1) button2StatePrev = 0;


    /* F1 PRESSED + START UNMUTES ALL TRACKS AND STARTS PLAYING*/
    if (button4State == 1 && button1State == 1 && button1StatePrev == 0 && writemode == 0) {
      for (i = 0; i <= (maxTracks - 1); i++) {
        pause[i] = 0;
      }
      lcd.setCursor(0, 1);
      lcd.print(F("all tracks play "));
      button1StatePrev = 1;
    }
    if (button2State == 0 && button2StatePrev == 1) button2StatePrev = 0;




    /* PLAY BUTTON IN STOP MODE*/

    if (button1State == 1 && button1StatePrev == 0 && runmode == 0 && button5State == 0) { //PLAY
      button1StatePrev = 1;
      if (sendclock >= 1) startFlag = 1; // send MIDI-Start command
      runmode = 1;
      for (i = 0; i <= (maxTracks - 1); i++) {
        //if (noteoffmode == 0) stepnumber[i] = -1;
        //if (noteoffmode > 0) stepnumber[i] = 0
        stepnumber[i] = 0;
        if (playonstart[i] == 0) {
          pause[i] = 0;
        }
        clockcounter[i] = 0;
      }
      clearscreen();

    }
    if (button1State == 0 && button1StatePrev == 1)  button1StatePrev = 0;

    /* PLAY BUTTON IN WRITE MODE */
    if (button1State == 1 && button1StatePrev == 0 && writemode == 1) { //PLAY
      button1StatePrev = 1;
      writemode = 0;
      for ( i = 0; i <= maxTracks; i++) {
        if (noteoffmode == 0) stepnumber[i] = -1;
        if (noteoffmode > 0) stepnumber[i] = 0;
        clockcounter[i] = 0;
      }
      runmode = 1;
      clearscreen();
      //if (button2State == 0) stepnumber[selectedtrack] = 0; // DO NOT RESET THE SEQUENCE IF LEAVING THE SETUP
    }
    if (button1State == 0 && button1StatePrev == 1)  button1StatePrev = 0;



    /* WRITE BUTTON IN STOP MODE */
    if (button3State == 1 && button3StatePrev == 0 && button1State == 0 && button2State == 0 && button4State == 0 && button5State == 0 && runmode == 0  && writemode == 0 && setupmode == 0) { //WRITE
      button3StatePrev = 1;
      writemode = 1;
      stepnumber[selectedtrack] = 0;
      clearscreen();
    }
    if (button3State == 0 && button3StatePrev == 1)  button3StatePrev = 0;

    /* STOP BUTTON IN STOP OR PLAY MODE*/
    if (button2State == 1 && button2StatePrev == 0 && button5State == 0) { //STOP
      button2StatePrev = 1;
      runmode = 0;
      clockInStatePrev = 0;
      triggerreset();
      for ( playedtrack = 0 ; playedtrack <= maxTracks - 1 ; playedtrack++ ) {
        if (tracktype[playedtrack] == 0 && noteOff[playedtrack] < 128 && noteOff[playedtrack] > 0) {
          MIDI.sendNoteOff(noteOff[playedtrack], velocity[playedtrack], outchannel[playedtrack]);
          noteOff[playedtrack] = 128;
        }
      }
      for (i = 0; i <= (maxTracks - 1); i++) {
        stepnumber[i] = 0;
        if (playonstart == 0) {
          pause[i] = 0;
        }
      }
    }
    if (button2State == 0 && button2StatePrev == 1) button2StatePrev = 0;




    /* ------------------------ STOP MODE STARTS HERE ------------------------ */
    if (runmode == 0 && writemode == 0) {
      lcd.setCursor(0, 0);
      lcd.print(F("STOP "));

      /* MIDITHRU IN STOP MODE*/
      if (MIDI.read()) {
        byte type = MIDI.getType();
        switch (type) {
          case midi::NoteOn:
            if (notethrumode > 0) MIDI.sendNoteOn(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
            break;

          case midi::NoteOff:
            if (notethrumode > 0) MIDI.sendNoteOff(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
            break;

          case midi::ControlChange:
            if ( controlthrumode == 1) MIDI.sendControlChange(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
            break;

          case midi::ProgramChange:
            if ( programthrumode == 1) MIDI.sendProgramChange(MIDI.getData1(), outchannel[selectedtrack]);
            break;

          case midi::PitchBend:
            if ( bendthrumode == 1) MIDI.sendPitchBend( (((MIDI.getData2() * 128) + MIDI.getData1() ) + 8192), outchannel[selectedtrack]);
            break;

          case midi::Start:
            if (clockmode > 0 ) {
              midiclockcounter = 0;
              lcd.setCursor(0, 0);
              /*
                for (i = 0; i <= (maxTracks - 1); i++) {
                if (noteoffmode == 0) stepnumber[i] = -1;
                if (noteoffmode > 0) stepnumber[i] = 0;
                byte tempvelocity = (center - ( random(1, 126) *  ((range + 1) / 128.0) ) / 2) + ((random(1, 126) *  ((range + 1) / 128) ));
                if (tempvelocity < 1) tempvelocity = 1;
                if (tempvelocity > 127) tempvelocity = 127;
                velocity[i] = tempvelocity;
                if (playonstart[i] == 0) {
                  pause[i] = 0;
                  waitEndOfPause[i] = 0;
                }
                }
              */
              runmode = 1;
              break;
            }
        }    /* END OF MIDITHRU IN STOP MODE */


      }
    }/* END OF STOP MODE */
    /* ------------------------ STOP MODE ENDS HERE --------------------------- */

    /* ------------------------ WRITE MODE STARTS HERE ------------------------ */

    if (runmode == 0 && writemode == 1) {
      lcd.setCursor(0, 0);
      lcd.print(F("WRITE"));
      lcd.print(F(" "));
      /*
        if (tracktype[selectedtrack] > 1) {
        lcd.print(F("+"));
        lcd.print(invalue[1]);
        lcd.print(F(" "));
        }
      */
      stepLCDposition = stepnumber[selectedtrack] % 16;

      /* WRITE NoteOn and NoteOff */
      if (MIDI.read()) {
        byte type = MIDI.getType()  ;
        switch (type) {
          case midi::NoteOn:
            if (MIDI.getChannel() == inchannel) {
              if (tracktype[selectedtrack] == 0 || tracktype[selectedtrack] == 3) {
                // MIDI THRU WHILE RECORD
                if (notethrumode > 1) MIDI.sendNoteOn(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
                // IF THIS TRACK IS A NOTE TRACK THEN WRITE A MIDI NOTE
                stepvalue[stepnumber[selectedtrack]][selectedtrack] = MIDI.getData1();

                lcd.setCursor(6, 0);
                lcd.print(F("#"));
                lcd.print(stepvalue[stepnumber[selectedtrack]][selectedtrack]);
                lcd.print(F(" "));
                lcd.setCursor(13, 0);
                lcd.print(stepnumber[selectedtrack] + 1);
                lcd.print(F(" "));
                printstep(stepLCDposition, 43); // Print a "+"
                stepnumber[selectedtrack]++;
                laststep[selectedtrack] = stepnumber[selectedtrack];
              }
            }
            break;
          case midi::NoteOff:
            if (notethrumode > 0) MIDI.sendNoteOff(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
            break;

          case midi::ControlChange:
            if ( controlthrumode == 1) MIDI.sendControlChange(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
            break;

          case midi::ProgramChange:
            if ( programthrumode == 1) MIDI.sendProgramChange(MIDI.getData1(), outchannel[selectedtrack]);
            break;

          case midi::PitchBend:
            if ( bendthrumode == 1) MIDI.sendPitchBend( (((MIDI.getData2() * 128) + MIDI.getData1() ) + 8192), outchannel[selectedtrack]);
            break;

        }
      }


      // IF THIS TRACK IS A CC TRACK OR VELOCITY TRACK THEN SELECT THE CV1 VALUE THAT WE WANT TO WRITE
      if (tracktype[selectedtrack] == 1 || tracktype[selectedtrack] == 2) {
        if (button4State == 1) {
          button4StatePrev = 1;
          lcd.setCursor(6, 0);
          lcd.print(F("#"));
          lcd.print(invalue[1]);
          lcd.print(F(" "));
        }
        if (button4State == 0 && button4StatePrev == 1) {
          button4StatePrev = 0;
          if (tracktype[selectedtrack] == 1 || tracktype[selectedtrack] == 2) {
            stepvalue[stepnumber[selectedtrack]][selectedtrack] = (invalue[1]);
          }
          lcd.setCursor(13, 0);
          lcd.print(stepnumber[selectedtrack] + 1);
          lcd.print(F(" "));
          printstep(stepLCDposition, 43); // Print a "+"
          stepnumber[selectedtrack]++;
          laststep[selectedtrack] = stepnumber[selectedtrack];
        }
      }
      // IF THIS TRACK IS A NOTE TRACK OR TRANSPOSE TRACK THEN WRITE A TIE
      if (tracktype[selectedtrack] == 0 || tracktype[selectedtrack] == 3) {
        if (button4State == 1 && button4StatePrev == 0) {
          button4StatePrev = 1;
        }
        if (button4State == 0 && button4StatePrev == 1) {
          button4StatePrev = 0;
          stepvalue[stepnumber[selectedtrack]][selectedtrack] = 129;
          lcd.setCursor(13, 0);
          lcd.print(stepnumber[selectedtrack] + 1);
          lcd.print(F(" "));
          printstep(stepLCDposition, 45); // Print a "-"
          stepnumber[selectedtrack]++;
          laststep[selectedtrack] = stepnumber[selectedtrack];
        }
      }

      /* Write a REST */
      if (button5State == 1  && button5StatePrev == 0) {
        stepvalue[stepnumber[selectedtrack]][selectedtrack] = 128;
        button5StatePrev = 1;
        //delay(2)
        lcd.setCursor(13, 0);
        lcd.print(stepnumber[selectedtrack] + 1);
        lcd.print(F(" "));
        printstep(stepLCDposition, 46); // Print a "."
        stepnumber[selectedtrack]++;
        laststep[selectedtrack] = stepnumber[selectedtrack];
      }
      if (button5State == 0 && button5StatePrev == 1) button5StatePrev = 0;

      /* One Step Backward with WRITE */
      if (button3State == 1 && button3StatePrev == 0 && writemode == 1 && stepnumber[selectedtrack] > 0) {
        button3StatePrev = 1;
        stepBack = 1;
      }
      if (button3State == 0 && button3StatePrev == 1) button3StatePrev = 0;

      if (stepBack == 1 ) {
        stepnumber[selectedtrack]--;
        laststep[selectedtrack] = stepnumber[selectedtrack];
        printstep(stepLCDposition - 1, 32); //print a " "
        lcd.setCursor(13, 0);
        lcd.print(stepnumber[selectedtrack] + 1);
        lcd.print(F(" "));
        stepBack = 0;
      }

      /* CONDITIONS FOR END WRITE */
      if (button2State == 1  && writemode == 1 && button3State == 0) { // END WRITE ON STOP
        button2StatePrev = 1;
        writemode = 0;
        runmode = 0;
        clockInStatePrev = 0;
        triggerreset();
        //midireset();
        clearscreen();
        lcd.setCursor(0, 0);
        lcd.print(F("STOP"));
      }
      if (button2State == 0 && button2StatePrev == 1) button2StatePrev = 0;

      if (button1State == 1  && writemode == 1) { // END WRITE ON START
        button1StatePrev = 1;
        writemode = 0;
        center = (invalue[1]);
        range = (invalue[2]);
        lcd.setCursor(0, 0);
        lcd.print(F("PLAY"));
        runmode = 1;
        clockInStatePrev = 0;
        triggerreset();
        //midireset();
        clearscreen();
        if (button1State == 0 && button1StatePrev == 1) button1StatePrev = 0;
      }

      if (stepnumber[selectedtrack] == maxSteps && writemode == 1 && button4StatePrev == 0 && button5StatePrev == 0) { // END WRITE IF MAX TRACKLENGTH REACHED
        writemode = 0;
        clockInStatePrev = 0;
        triggerreset();
        //midireset();
        clearscreen();
      }
    }
    /* ------------------------ WRITE MODE ENDS HERE ------------------------ */
  } /* END OF SETUPMODE == 0*/


  /* ------------------------ PLAY MODE STARTS HERE ------------------------ */
  if (runmode == 1 && writemode == 0) {

    /* SWITCHING THE PAUSE MODE */
    /* START WHILE PLAY MODE ACTIVATES PAUSE */
    if (button1State == 1 && button1StatePrev == 0 && pause[selectedtrack] == 0 && setupmode == 0) { // ENTER PAUSE MODE
      button1StatePrev = 1;
      pause[selectedtrack] = 1;
    }
    /* START WHILE PAUSE MODE SETS TO WAIT-END-OF-PAUSE */
    if (button1State == 1 && button1StatePrev == 0 && pause[selectedtrack] == 1 && setupmode == 0) { // ENTER WAIT PAUSE END MODE
      button1StatePrev = 1;
      pause[selectedtrack] = 2;
    }
    /* START WHILE WAIT-END-OF-PAUSE ACTIVATES PLAY IMMEDIATELY */
    if (button1State == 1 && button1StatePrev == 0  && pause[selectedtrack] == 2 && setupmode == 0) {
      pause[selectedtrack] = 0;
      button1StatePrev = 1;
    }
    /* STEP=0 WHILE WAIT-END-OF-PAUSE ACTIVATES PLAY AGAIN*/
    if (stepnumber[selectedtrack] == 0 && pause[selectedtrack] == 2) {
      pause[selectedtrack] = 0;
    }
    if (button1State == 0 && button1StatePrev == 1) {
      button1StatePrev = 0;
    }



    /* READ INCOMING MIDI CLOCK */
    if (MIDI.read()) {
      byte type = MIDI.getType();
      switch (type) {
        case midi::Clock:
          if (clockmode == 1) {
            if (midiclockcounter == 0) {
              if (clockmode == 1) digitalWrite(clockOutPin, HIGH);
            }

            if (midiclockcounter == ( 1 * clockfactor)) {
              if (clockmode == 1) digitalWrite(clockOutPin, LOW);
            }

            midiclockcounter++;
            if (midiclockcounter == (2 * clockfactor)) {
              midiclockcounter = 0;
            }
          }
          break;

        case midi::Stop:
          midiclockcounter = 0;
          runmode = 0;
          for ( playedtrack = 0 ; playedtrack < maxTracks ; playedtrack++ ) {
            if (tracktype[playedtrack] == 0 && noteOff[playedtrack] < 128 && noteOff[playedtrack] > 0) {
              MIDI.sendNoteOff(noteOff[playedtrack], velocity[playedtrack], outchannel[playedtrack]);
              noteOff[playedtrack] = 128;
            }
          }

          break;

        case midi::NoteOn:
          if ((notethrumode == 2 || notethrumode == 3) && (transposemode == 0 || (transposemode == 3 && button5State == 0) || transposemode > 3)) MIDI.sendNoteOn(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
          if (transposemode == 1 || transposemode == 2) {
            transposeall = MIDI.getData1();
            transposeall = transposeall - transposecenter;
            if (tracktype[selectedtrack] == 0) showtranspose(transposeall);
          }
          if (transposemode == 3 && button5State == 1) {
            transposeall = MIDI.getData1();
            transposeall = transposeall - transposecenter;
            if (tracktype[selectedtrack] == 0) showtranspose(transposeall);
          }


          for (i = 0; i < maxTracks; i++) {
            transpose[i] = transposeall;
          }
          break;

        case midi::NoteOff:
          if (notethrumode > 0) MIDI.sendNoteOff(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
          if (transposemode == 2) transposeall = 0;
          break;

        case midi::ControlChange:
          if ( controlthrumode == 1) MIDI.sendControlChange(MIDI.getData1(), MIDI.getData2(), outchannel[selectedtrack]);
          break;

        case midi::ProgramChange:
          if ( programthrumode == 1) MIDI.sendProgramChange(MIDI.getData1(), outchannel[selectedtrack]);
          break;

        case midi::PitchBend:
          if ( bendthrumode == 1) MIDI.sendPitchBend( (((MIDI.getData2() * 128) + MIDI.getData1() ) + 8192), outchannel[selectedtrack]);
          break;



      }
    } //END OF MIDI-READ




    clockInState = !digitalRead(clockInPin);
    if (clockInState == 1 && clockInStatePrev == 0) {
      for ( playedtrack = 0 ; playedtrack <= maxTracks - 1 ; playedtrack++ ) {
        if ((clockdivider[playedtrack] - clockcounter[playedtrack]) % clockdivider[playedtrack] == 0) {
          ontrigger[playedtrack] = 1;
        }
        clockcounter[playedtrack]++;
      }
      clockInStatePrev = 1;
    }
    if (clockInState == 0 && clockInStatePrev == 1) {
      for ( playedtrack = 0 ; playedtrack <= maxTracks - 1 ; playedtrack++ ) {
        if ((clockdivider[playedtrack] - clockcounter[playedtrack]) % clockdivider[playedtrack] == 0) {
          offtrigger[playedtrack] = 1;
          clockcounter[playedtrack] = 0;
        }
      }
      clockInStatePrev = 0;
    }


    // First we prepare transpose and velocity data for the current step
    for ( playedtrack = 0 ; playedtrack < maxTracks ; playedtrack++ ) {
      if (ontrigger[playedtrack] == 1 && tracktype[playedtrack] >= 2) currentvalue[playedtrack] = stepvalue[stepnumber[playedtrack]][playedtrack];
    }


    // Then we prepare the note data from velocity and transpose
    center = (invalue[1]);
    range = (invalue[2]);
    if (transposemode > 3) {
      if (currentvalue[transposemode - 4] < 128) transposeall = currentvalue[transposemode - 4]  - transposecenter;
      if (tracktype[selectedtrack] == 0) showtranspose(transposeall);
    }
    if (velocitymode > 2) {
      if (currentvalue[velocitymode - 3] < 128) velocityall = currentvalue[velocitymode - 3] + 1;
    }
    for ( playedtrack = 0 ; playedtrack < maxTracks ; playedtrack++ ) {
      if (ontrigger[playedtrack] == 1 && tracktype[playedtrack] == 0) {
        if (velocitysource[playedtrack] < maxTracks) {
          byte tempvelocity = (currentvalue[velocitysource[playedtrack]] * (range / 128.0)) + center;
          if (tempvelocity < 1) tempvelocity = 1;
          if (tempvelocity > 127) tempvelocity = 127;
          velocity[playedtrack] = tempvelocity;
        }
        if (velocitysource[playedtrack] == maxTracks) velocity[playedtrack] = center;
        if (velocitysource[playedtrack] == maxTracks + 1) {
          byte tempvelocity = (random(1, 126) * (range / 128.0)) + ((2 * center) / 3);
          if (tempvelocity < 1) tempvelocity = 1;
          if (tempvelocity > 127) tempvelocity = 127;
          velocity[playedtrack] = tempvelocity;
        }

        if (transposesource[playedtrack] < maxTracks && transposesource[playedtrack] != playedtrack) {
          if (currentvalue[transposesource[playedtrack]] < 128 && currentvalue[transposesource[playedtrack]] > 0) transpose[playedtrack] = currentvalue[transposesource[playedtrack]] - transposecenter;
        }
        if (transposesource[playedtrack] == maxTracks) transpose[playedtrack] = 0;
        if (transposesource[playedtrack] == maxTracks + 1) transpose[playedtrack] = transposeall;
      }
    }

    // We send all ControlChange Date befor we send the notes, so the sound changes get in effect, before the notes invoke the sound
    for ( playedtrack = 0 ; playedtrack <= maxTracks - 1 ; playedtrack++ ) {
      if (ontrigger[playedtrack] == 1) {
        if (tracktype[playedtrack] == 1  && pause[playedtrack] == 0 && stepvalue[stepnumber[playedtrack]][playedtrack] < 128) MIDI.sendControlChange(ccnumber[playedtrack], stepvalue[stepnumber[playedtrack]][playedtrack], outchannel[playedtrack]);
      }
    }

    for ( playedtrack = 0 ; playedtrack <= maxTracks - 1 ; playedtrack++ ) {
      if (ontrigger[playedtrack] == 1) {
        if (noteoffmode == 0 && stepvalue[stepnumber[playedtrack]][playedtrack] <= 128 && noteOff[playedtrack] > 0)  {
          MIDI.sendNoteOff(noteOff[playedtrack], velocity[playedtrack], outchannel[playedtrack]);
          noteOff[playedtrack] = 0;
        }
        //transpose[playedtrack] = 0;
        //lcd.setCursor(5, 0);
        //lcd.print(transpose[selectedtrack]);
        //lcd.print(transposecenter);
        //lcd.setCursor(9, 0);
        //lcd.print(stepvalue[stepnumber[selectedtrack]][selectedtrack]);
        if (tracktype[playedtrack] == 0  && pause[playedtrack] == 0 && stepvalue[stepnumber[playedtrack]][playedtrack] < 128) MIDI.sendNoteOn(stepvalue[stepnumber[playedtrack]][playedtrack] + transpose[playedtrack], velocity[playedtrack], outchannel[playedtrack]);
        if (tracktype[playedtrack] == 0  && pause[playedtrack] == 0 && stepvalue[stepnumber[playedtrack]][playedtrack] < 128) noteOff[playedtrack] = stepvalue[stepnumber[playedtrack]][playedtrack] + transpose[playedtrack];
        ontrigger[playedtrack] = 0;
      }
    }




    /*PRINT THE STEP SYMBOLS*/
    stepLCDposition = stepnumber[selectedtrack] % 16;
    if (stepvalue[stepnumber[selectedtrack]][selectedtrack] < 128) {
      printstep(stepLCDposition, 43); //print a "+"
    }
    if (stepvalue[stepnumber[selectedtrack]][selectedtrack] == 128) {
      printstep(stepLCDposition, 46); //print a "."
    }
    if (stepvalue[stepnumber[selectedtrack]][selectedtrack] == 129) {
      printstep(stepLCDposition, 45); //print a "-"
    }







    for ( playedtrack = 0 ; playedtrack <= maxTracks - 1 ; playedtrack++ ) {
      if (offtrigger[playedtrack] == 1) {
        if (tracktype[playedtrack] == 0 && noteOff[playedtrack] < 128 && noteOff[playedtrack] > 0 && noteoffmode == 1) {
          MIDI.sendNoteOff(noteOff[playedtrack], velocity[playedtrack], outchannel[playedtrack]);
          noteOff[playedtrack] = 0;
        }
        offtrigger[playedtrack] = 0;
        stepnumber[playedtrack]++;
        if (stepnumber[playedtrack] == (laststep[playedtrack])) stepnumber[playedtrack] = 0;
      }
    }
  } //END OF RUNMODE
  /* ------------------------ PLAY MODE ENDS HERE ------------------------ */


  /*========HERE BEGINS THE SETUP SECTION========*/
  if (setupmode == 1 ) { //SETUP

    MIDI.read();
    lcd.setCursor(0, 1);
    if (edit == 0) lcd.print(F(" "));
    if (edit == 1) lcd.print(F("!"));

    /* STOP+PLAY exits the setup */
    if ( button1State == 1 && button2State == 1 && edit == 0) {
      setupmode = 0;
      clearscreen();
    }

    /* START  PRESSED + BUTTON F1 DECREASE SELECTED TRACK */
    if (button1State == 1 && button4State == 1 && button4StatePrev == 0 && saveMulti == 1) {
      button4StatePrev = 1;
      selectedtrack--;
      if (selectedtrack == 255) selectedtrack = maxTracks - 1;
    }
    if (button4State == 0 && button4StatePrev == 1) button4StatePrev = 0;

    /* START PRESSED + BUTTON F2 INCREASES SELECTED TRACK */
    if (button1State == 1 && button5State == 1 && button5StatePrev == 0 && writemode == 0 && saveMulti == 1) {
      button5StatePrev = 1;
      selectedtrack++;
      if (selectedtrack == maxTracks) selectedtrack = 0;
    }
    if (button5State == 0 && button5StatePrev == 1) button5StatePrev = 0;


    /*
      Stepping thru the setup menu
    */
    if (edit == 0) {
      menuPrev = menu;
      menu = getDataValue(menu, 25, 0, 0, 1);
      if (menu != menuPrev) clearscreen();

      if (button3State == 1 && button3StatePrev == 0) { // ENTER EDIT MODE
        button3StatePrev = 1;
        if (menu == 15 && runmode == 1) {
          lcd.setCursor(1, 1);
          lcd.print(F("Not while Play"));
        } else {
          edit = 1;
        }
      }
    }

    do_menu();

    if (edit == 1) {
      if (button3State == 1 && button3StatePrev == 0 && ( menu < 16 || menu > 20 )) {
        button3StatePrev = 1;
        edit = 0;
        if (saveMulti == 0) EEPROM.update(saveMemory, saveValue);
        if (saveMulti == 1) {
          for (i = 0; i < maxTracks; i++) {
            EEPROM.update(saveMemory + i, saveValueMulti[i]);
          }
        }
      }
    }

    if (button3State == 0 && button3StatePrev == 1) {
      button3StatePrev = 0;
    }

    /* AGAIN SOME DEBUGGING OUTPUT
      lcd.setCursor(9, 0);
      lcd.print(button3State);
      lcd.print(button3StatePrev);
      lcd.print(edit);
      /**/
  }


}
/* END OF LOOP */
